/*
 * Copyright (c) 2022-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <fstream>

#include <infiniband/misc/regexp/regexp.h>
#include <infiniband/ibdiag/ibdiag_ppcc.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>

const string PPCCAlgoDatabase::sectionStartStr = "ca_algo_start";
const string PPCCAlgoDatabase::sectionEndStr = "ca_algo_end";
const string PPCCAlgoDatabase::nameStr = "name";
const string PPCCAlgoDatabase::versionStr = "version";
const string PPCCAlgoDatabase::releaseDateStr = "algo_release_date";
const string PPCCAlgoDatabase::descriptionStr = "algo_description";
const string PPCCAlgoDatabase::authorStr = "algo_author";
const string PPCCAlgoDatabase::supportedDevicesStr = "supported_devices";
const string PPCCAlgoDatabase::ppccParameterNameListStr = "ppcc_parameter_name_list";
const string PPCCAlgoDatabase::ppccParameterMinValsStr = "ppcc_parameter_min_vals";
const string PPCCAlgoDatabase::ppccParameterMaxValsStr = "ppcc_parameter_max_vals";
const string PPCCAlgoDatabase::ppccCountersNameListStr = "ppcc_counters_name_list";

int PPCCAlgoDatabase::ParseUint32(const char *str, size_t line, uint32_t &val)
{
    char *endptr = NULL;

    val = (uint32_t)strtoul(str, &endptr, 0);

    if (*str && *endptr == '\0')
        return 0;

    ERR_PRINT("Failed to parse integer \"%s\" at line %lu\n", str, line);
    return 1;
}

int PPCCAlgoDatabase::ParseTupleList(const char *str, vector<pair<string, string>> &tupleList)
{
    regExp tupleListRe("\\[\\([^,]+,[^,]+\\)(,\\([^,]+,[^,]+\\))*\\]");
    regExp tupleRe("\\(([^,]+),([^,]+)\\)");

    rexMatch *match = tupleListRe.apply(str);
    if (!match)
        return 1;

    delete match;

    str++; // remove first '['

    while ((match = tupleRe.apply(str))) {
        tupleList.push_back({match->field(1), match->field(2)});
        str += match->field(0).length() + 1;
        delete match;
    }

    return 0;
}

int PPCCAlgoDatabase::ParseSimpleList(const char *str, vector<string> &simpleList)
{
    regExp listRe("\\[[^,]+(,[^,]+)*\\]");
    regExp itemRe("([^,]+)(,|])");

    rexMatch *match = listRe.apply(str);
    if (!match)
        return 1;

    delete match;

    while ((match = itemRe.apply(str + 1))) {
        simpleList.push_back(match->field(1));
        str += match->field(0).length();
        delete match;
    }

    return 0;
}

int PPCCAlgoDatabase::HandleKeyVal(const string key, const string &val, size_t line,
                                   ParserPPCCAlgo &parserAlgo)
{
    int rc = 0;

    vector<pair<string, string>> tupleList;
    vector<string> simpleList;

    if (key == nameStr) {

        if (val.empty())
            WARN_PRINT("No value for 'name' at line %lu\n", line);

        parserAlgo.name = val;

    } else if (key == versionStr) {
        regExp versionRe("^([0-9]+)\\.([0-9]+)\\.([0-9]+)$");

        rexMatch *match = versionRe.apply(val.c_str());

        do {
            if (!match) {
                rc = 1;
                break;
            }

            uint32_t algoId = 0, versionMajor = 0, versionMinor = 0;

            if ((rc = ParseUint32(match->field(1).c_str(), line, algoId)))
                break;

            if ((rc = ParseUint32(match->field(2).c_str(), line, versionMajor)))
                break;

            if ((rc = ParseUint32(match->field(3).c_str(), line, versionMinor)))
                break;

            parserAlgo.versionHash = ALGO_VERSION_HASH_MAKE(algoId, versionMajor, versionMinor);
        } while(0);

        if (match)
            delete match;

    } else if (key == releaseDateStr) {
        // not used

    } else if (key == descriptionStr) {
        // not used

    } else if (key == authorStr) {
        // not used

    } else if (key == supportedDevicesStr) {
        // not used

    } else if (key == ppccParameterNameListStr) {
        if (!(rc = ParseTupleList(val.c_str(), tupleList))) {
            for (vector<pair<string, string>>::iterator it = tupleList.begin();
                    it != tupleList.end(); ++it) {
                uint32_t defVal = 0;
                if ((rc = ParseUint32(it->second.c_str(), line, defVal)))
                    break;
                parserAlgo.paramDefVals.push_back({it->first, defVal});
            }
        }
    } else if (key == ppccParameterMinValsStr) {
        if (!(rc = ParseTupleList(val.c_str(), tupleList))) {
            for (vector<pair<string, string>>::iterator it = tupleList.begin();
                    it != tupleList.end(); ++it) {
                uint32_t minVal = 0;
                if ((rc = ParseUint32(it->second.c_str(), line, minVal)))
                    break;
                parserAlgo.paramMinVals.push_back({it->first, minVal});
            }
        }
    } else if (key == ppccParameterMaxValsStr) {
        if (!(rc = ParseTupleList(val.c_str(), tupleList))) {
            for (vector<pair<string, string>>::iterator it = tupleList.begin();
                    it != tupleList.end(); ++it) {
                uint32_t maxVal = 0;
                if ((rc = ParseUint32(it->second.c_str(), line, maxVal)))
                    break;
                parserAlgo.paramMaxVals.push_back({it->first, maxVal});
            }
        }
    } else if (key == ppccCountersNameListStr) {
        if (!(rc = ParseSimpleList(val.c_str(), simpleList))) {
            for (vector<string>::iterator it = simpleList.begin();
                    it != simpleList.end(); ++it) {
                parserAlgo.counters.push_back(*it);
            }
        }
    } else { // Unknown key
        WARN_PRINT("Unknown key '%s' at line %lu\n", key.c_str(), line);
    }

    if (rc)
        ERR_PRINT("Failed to parse %s value at line %lu\n", key.c_str(), line);

    return rc;
}

int PPCCAlgoDatabase::HandleLine(const string &key, const string &val,
                                 ParserGlobalState &globalState, size_t line,
                                 ParserPPCCAlgo &parserAlgo)
{
    if (key.empty())
        return 0;

    if (key == sectionStartStr && val.empty()) {
        if (globalState != PARSE_GLOBAL) {
            ERR_PRINT("Unexpected %s at line %lu\n", sectionStartStr.c_str(), line);
            return 1;
        }

        parserAlgo = {};
        parserAlgo.startLine = line;

        globalState = PARSE_SECTION;
        return 0;
    }

    if (key == sectionEndStr && val.empty()) {
        if (globalState != PARSE_SECTION) {
            ERR_PRINT("Unexpected %s at line %lu\n", sectionEndStr.c_str(), line);
            return 1;
        }

        parserAlgo.endLine = line;
        if (FillNewAlgo(parserAlgo))
            return 1;

        globalState = PARSE_GLOBAL;
        return 0;
    }

    if (HandleKeyVal(key, val, line, parserAlgo))
        return 1;

    return 0;
}

int PPCCAlgoDatabase::FillNewAlgo(ParserPPCCAlgo &parserAlgo)
{
    if (parserAlgo.versionHash == 0) {
        ERR_PRINT("Algo [%lu-%lu] doesn't have version hash\n", parserAlgo.startLine,
                  parserAlgo.endLine);
        return 1;
    }

    if (algoMap.find(parserAlgo.versionHash) != algoMap.end()) {
        ERR_PRINT("Algo [%lu-%lu] - algo with version hash 0x%08X already exists\n",
                  parserAlgo.startLine, parserAlgo.endLine, parserAlgo.versionHash);
        return 1;
    }

    PPCCAlgo &algo = algoMap.insert({parserAlgo.versionHash,
                                     {parserAlgo.name, parserAlgo.versionHash}}).first->second;

    for (vector<pair<string, uint32_t>>::iterator it = parserAlgo.paramDefVals.begin();
            it != parserAlgo.paramDefVals.end(); ++it) {
        if (algo.GetParameter(it->first)) {
            ERR_PRINT("Algo [%lu-%lu] - parameter with name %s already exists\n",
                      parserAlgo.startLine, parserAlgo.endLine, it->first.c_str());
            return 1;
        }

        algo.AddParameter(it->first, it->second);
    }

    for (vector<pair<string, uint32_t>>::iterator it = parserAlgo.paramMinVals.begin();
            it != parserAlgo.paramMinVals.end(); ++it) {
        PPCCParameter *param;

        if (!(param = algo.GetParameter(it->first))) {
            ERR_PRINT("Algo [%lu-%lu] - parameter with name %s doesn't exist (min value)\n",
                      parserAlgo.startLine, parserAlgo.endLine, it->first.c_str());
            return 1;
        }

        param->SetMinVal(it->second);
    }

    for (vector<pair<string, uint32_t>>::iterator it = parserAlgo.paramMaxVals.begin();
            it != parserAlgo.paramMaxVals.end(); ++it) {
        PPCCParameter *param;

        if (!(param = algo.GetParameter(it->first))) {
            ERR_PRINT("Algo [%lu-%lu] - parameter with name %s doesn't exist (max value)\n",
                      parserAlgo.startLine, parserAlgo.endLine, it->first.c_str());
            return 1;
        }

        param->SetMaxVal(it->second);
    }

    for (vector<string>::iterator it = parserAlgo.counters.begin();
            it != parserAlgo.counters.end(); ++it) {
        algo.AddCounter(*it);
    }

    return 0;
}

int PPCCAlgoDatabase::ParseDir(const string &ppcc_path)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    string ppcc_pattern = ppcc_path;
    struct stat sb;

    if ( stat(ppcc_path.c_str(), &sb) == 0 ) {
        if ((sb.st_mode & S_IFMT) == S_IFDIR) // directory
            ppcc_pattern += "/*";
    }

    vector<string> files = IBFabric::getFilesByPattern(ppcc_pattern);

    if (files.empty()) {
        WARN_PRINT("Failed to find file(s) in \"%s\"\n", ppcc_path.c_str());
    } else {
        for (const auto& file : files)
            ParseFile(file);
    }

    IBDIAG_RETURN(rc);
}

int PPCCAlgoDatabase::ParseFile(const string &fileName)
{
    ifstream stream(fileName);

    if (!stream.is_open()) {
        int e = errno;
        ERR_PRINT("Failed to open file \"%s\" - %s\n", fileName.c_str(), strerror(e));
        return 1;
    }

    INFO_PRINT("Parsing PPCC algo file: '%s'...\n", fileName.c_str());

    size_t prev_size = algoMap.size();

    ParserPPCCAlgo parserAlgo;

    char c;
    size_t line = 1;

    string key;
    string val;

    ParserGlobalState globalState = PARSE_GLOBAL;
    ParserSectionState sectionState = PARSE_KEY;

    while (true) {
        stream >> noskipws >> c;

        if (c == '#') {
            while ((stream >> noskipws >> c) && (c != '\n'))
                ;
        }

        if (c == '\n' || !stream) {
            // Process key / value found on line
            if (HandleLine(key, val, globalState, line, parserAlgo))
                return 1;

            // Reset the line
            sectionState = PARSE_KEY;

            key.clear();
            val.clear();

            line++;

            if (!stream)
                break;

            continue;
        }

        if (isspace(c))
            continue;

        if (sectionState == PARSE_KEY) {
            if (c == ':')
                sectionState = PARSE_VALUE;
            else
                key += c;
        } else if (sectionState == PARSE_VALUE) {
            val += c;
        }
    }

    // File must end in PARSE_GLOBAL
    if (globalState != PARSE_GLOBAL) {
        ERR_PRINT("Found %s without %s, line %lu\n",
                  sectionStartStr.c_str(),
                  sectionEndStr.c_str(),
                  line);
        return 1;
    }

    if (prev_size == algoMap.size())
        WARN_PRINT("Failed to find new Algorithms in file: \"%s\"\n", fileName.c_str());

    return 0;
}

PPCCAlgo *PPCCAlgoDatabase::GetAlgo(uint16_t algoId, uint8_t versionMajor, uint8_t versionMinor)
{
    uint32_t versionHash = ALGO_VERSION_HASH_MAKE(algoId, versionMajor, versionMinor);

    PPCCAlgo *algo = NULL;

    for (map<uint32_t, PPCCAlgo>::iterator it = algoMap.begin();
            it != algoMap.end(); ++it) {
        PPCCAlgo &itAlgo = it->second;

        uint16_t itAlgoId = ALGO_VERSION_HASH_ALGO_ID(itAlgo.GetVersionHash());
        uint8_t itVersionMajor = ALGO_VERSION_HASH_VERSION_MAJOR(itAlgo.GetVersionHash());
        uint8_t itVersionMinor = ALGO_VERSION_HASH_VERSION_MINOR(itAlgo.GetVersionHash());

        if (itAlgoId == algoId && itVersionMajor == versionMajor &&
                itVersionMinor >= versionMinor) {
            algo = &itAlgo;
            break;
        }

        if (itAlgo.GetVersionHash() > versionHash)
            break;
    }

    return algo;
}
