/*
 * SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: LicenseRef-NvidiaProprietary
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */
#include <string>
#include <cstring>
#include <cctype>
#include <cstdint>
#include <cstdlib>


#include "cable_db_csv_record.h"
#include <infiniband/ibdiag/ibdiag_csv_parser.h>
#include <infiniband/ibdiag/ibdiag.h>




#define ADD_CSV_FIELD(T, name, member)                                                      \
    parse_section_info.push_back(                                                           \
            ParseFieldInfo<T>(name,                                                         \
                    [](T & obj, const char * value)                                         \
                        { return Parse(value, obj.member); } ))

#define ADD_CSV_FIELD_CHAR(T, name, member)                                                 \
    parse_section_info.push_back(                                                           \
            ParseFieldInfo<T>(name,                                                         \
                    [](T & obj, const char * value) {                                       \
                     std::string str;                                                       \
                        bool rc = Parse(value, str);                                        \
                        if (rc) {                                                           \
                         CLEAR_STRUCT(obj.member);                                          \
                            strncpy((char *)obj.member, str.c_str(),                        \
                                    sizeof(obj.member) - 1);                                \
                        }                                                                   \
                        return rc;                                                          \
                    } ));


//
// Section ROW_MODULE_INFO
//  - pddr_module_info
//
int ModuleInfoRecord::Init(vector < ParseFieldInfo <class ModuleInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(ModuleInfoRecord, "ID", m_id);
    ADD_CSV_FIELD(ModuleInfoRecord, "Technology", m_technology);
    ADD_CSV_FIELD(ModuleInfoRecord, "Side",  m_side);

    #define __ADD__(x)  ADD_CSV_FIELD(ModuleInfoRecord, #x, m_data.x)
    __ADD__(cable_technology);
    __ADD__(cable_breakout);
    __ADD__(ext_ethernet_compliance_code);
    __ADD__(ethernet_compliance_code);
    __ADD__(cable_type);
    __ADD__(cable_vendor);
    __ADD__(cable_length);
    __ADD__(cable_identifier);
    __ADD__(cable_power_class);
    __ADD__(max_power);
    __ADD__(cable_rx_amp);
    __ADD__(cable_rx_emphasis);
    __ADD__(cable_tx_equalization);
    __ADD__(cable_attenuation_25g);
    __ADD__(cable_attenuation_12g);
    __ADD__(cable_attenuation_7g);
    __ADD__(cable_attenuation_5g);
    __ADD__(cable_rx_post_emphasis);
    __ADD__(rx_cdr_cap);
    __ADD__(tx_cdr_cap);
    __ADD__(rx_cdr_state);
    __ADD__(tx_cdr_state);
    __ADD__(fw_version);
    __ADD__(temperature);
    __ADD__(voltage);
    __ADD__(rx_power_lane0);
    __ADD__(rx_power_lane1);
    __ADD__(rx_power_lane2);
    __ADD__(rx_power_lane3);
    __ADD__(rx_power_lane4);
    __ADD__(rx_power_lane5);
    __ADD__(rx_power_lane6);
    __ADD__(rx_power_lane7);
    __ADD__(tx_power_lane0);
    __ADD__(tx_power_lane1);
    __ADD__(tx_power_lane2);
    __ADD__(tx_power_lane3);
    __ADD__(tx_power_lane4);
    __ADD__(tx_power_lane5);
    __ADD__(tx_power_lane6);
    __ADD__(tx_power_lane7);
    __ADD__(tx_bias_lane0);
    __ADD__(tx_bias_lane1);
    __ADD__(tx_bias_lane2);
    __ADD__(tx_bias_lane3);
    __ADD__(tx_bias_lane4);
    __ADD__(tx_bias_lane5);
    __ADD__(tx_bias_lane6);
    __ADD__(tx_bias_lane7);
    __ADD__(temperature_high_th);
    __ADD__(voltage_high_th);
    __ADD__(voltage_low_th);
    __ADD__(rx_power_high_th);
    __ADD__(rx_power_low_th);
    __ADD__(tx_power_high_th);
    __ADD__(tx_power_low_th);
    __ADD__(tx_bias_high_th);
    __ADD__(tx_bias_low_th);
    __ADD__(module_st);
    __ADD__(rx_power_type);
    __ADD__(did_cap);
    __ADD__(smf_length);
    __ADD__(wavelength);
    __ADD__(active_set_host_compliance_code);
    __ADD__(active_set_media_compliance_code);
    __ADD__(ib_compliance_code);
    __ADD__(nbr250);
    __ADD__(nbr100);
    __ADD__(monitor_cap_mask);
    __ADD__(ib_width);
    __ADD__(dp_st_lane0);
    __ADD__(dp_st_lane1);
    __ADD__(dp_st_lane2);
    __ADD__(dp_st_lane3);
    __ADD__(dp_st_lane4);
    __ADD__(dp_st_lane5);
    __ADD__(dp_st_lane6);
    __ADD__(dp_st_lane7);
    __ADD__(length_om2);
    __ADD__(length_om3);
    __ADD__(length_om4);
    __ADD__(length_om5);
    __ADD__(length_om1);
    __ADD__(wavelength_tolerance);
    __ADD__(memory_map_rev);
    __ADD__(memory_map_compliance);
    __ADD__(date_code);
    __ADD__(connector_type);
    __ADD__(vendor_oui);
    __ADD__(cable_attenuation_53g);
    __ADD__(rx_output_valid);
    __ADD__(tx_input_freq_sync);
    __ADD__(error_code);
    __ADD__(cdr_vendor);
    __ADD__(max_fiber_length);

    #undef __ADD__

 #define ADD_CHAR(name, member) ADD_CSV_FIELD_CHAR(ModuleInfoRecord, name, m_data.member);

    ADD_CHAR("vendor_name", vendor_name);
    ADD_CHAR("vendor_pn", vendor_pn);
    ADD_CHAR("revision",vendor_rev);
    ADD_CHAR("vendor_sn", vendor_sn);
 #undef ADD_CHAR


    return 0;
}


//
// Section ROW_LATCHED_FLAG_INFO
//  - module_latched_flag_info
//
int LatchedFlagInfoRecord::Init(vector < ParseFieldInfo <class LatchedFlagInfoRecord> > &parse_section_info)
{
    ADD_CSV_FIELD(LatchedFlagInfoRecord, "ID", m_id);
    ADD_CSV_FIELD(LatchedFlagInfoRecord, "Technology", m_technology);
    ADD_CSV_FIELD(LatchedFlagInfoRecord, "Side",  m_side);

    #define __ADD__(x)  ADD_CSV_FIELD(LatchedFlagInfoRecord, #x, m_data.x)

    __ADD__(dp_fw_fault);
    __ADD__(mod_fw_fault);
    __ADD__(vcc_flags);
    __ADD__(temp_flags);
    __ADD__(tx_ad_eq_fault);
    __ADD__(tx_cdr_lol);
    __ADD__(tx_los);
    __ADD__(tx_fault);
    __ADD__(tx_power_lo_war);
    __ADD__(tx_power_hi_war);
    __ADD__(tx_power_lo_al);
    __ADD__(tx_power_hi_al);
    __ADD__(tx_bias_lo_war);
    __ADD__(tx_bias_hi_war);
    __ADD__(tx_bias_lo_al);
    __ADD__(tx_bias_hi_al);
    __ADD__(rx_cdr_lol);
    __ADD__(rx_los);
    __ADD__(rx_power_lo_war);
    __ADD__(rx_power_hi_war);
    __ADD__(rx_power_lo_al);
    __ADD__(rx_power_hi_al);
    __ADD__(rx_output_valid_change);
    //__ADD__(rx_input_valid_change);

    #undef __ADD__

    return 0;
}
