 /*
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#pragma once

#include <stdint.h>
#include <vector>
#include <map>
#include <string>
#include <ostream>

#include "infiniband/ibutils/release_containers.h"

#include "uphy_common.h"
#include "uphy_version.h"
#include "uphy_dataset_key.h"


namespace UPHY
{
    class DataSet
    {
        public:
            class Enumerator
            {
                public:
                    class Label
                    {
                        public:
                            typedef uint8_t value_t;

                        private:
                            const std::string   m_name;
                            const value_t       m_value;

                        public:
                            inline const std::string &  name() const    { return m_name;    }
                            value_t                     value() const   { return m_value;   }

                        private:
                            inline Label() : m_value(0) {}

                        public:
                            inline Label(const std::string & name, value_t & value)
                                : m_name(name), m_value(value) {}
                    };

                public:
                    typedef std::map<Label::value_t, const Label*> labels_collection_t;

                public:
                    typedef uint8_t width_t;

                private:
                    const std::string       m_name;
                    const width_t           m_width;

                private:
                    labels_collection_t     m_labels;

                public:
                    inline const std::string & name() const    { return m_name;    }
                    inline const width_t &     width() const   { return m_width;   }

                public:
                    inline const labels_collection_t & labels() const { return m_labels; }

                private:
                    inline Enumerator() : m_width(0) {}

                public:
                    inline Enumerator(const std::string & name, width_t width)
                        : m_name(name), m_width(width)
                    {}

                public:
                    ~Enumerator() {
                        release_container_data(m_labels);
                    }

                public:
                    const Label * add(const Label * label)
                    {
                        if( label == nullptr )
                            return nullptr;

                        auto x = m_labels.insert(std::make_pair(label->value(), label));

                        return x.second ? label : nullptr;
                    }

                public:
                    const Label * get(Label::value_t value) const
                    {
                        auto x = m_labels.find( value );
                        return x != m_labels.end() ? x->second : nullptr;
                    }

            };

        public:
            class Register
            {
                public:
                    typedef uint16_t address_t;

                public:
                    class Field
                    {
                        public:
                            typedef uint16_t range_t;
                            typedef uint16_t width_t;

                        private:
                            const std::string           m_name;
                            const width_t               m_width;
                            const range_t               m_start;
                            const range_t               m_end;
                            const register_field_t      m_type;
                            const Enumerator *          m_enumerator;

                        public:
                            inline const std::string &  name() const        { return m_name;        }
                            inline width_t              width() const       { return m_width;       }
                            inline range_t              start() const       { return m_start;       }
                            inline range_t              end() const         { return m_end;         }
                            inline register_field_t     type() const        { return m_type;        }
                            inline const Enumerator *   enumerator() const  { return m_enumerator;  }

                        private:
                            inline Field()
                                : m_width(0), m_start(0), m_end(0), m_type(RegisterFieldType_None), m_enumerator(nullptr)
                            {}

                        public:
                            inline Field( const std::string &   name,
                                            width_t             width,
                                            range_t             start,
                                            range_t             end,
                                            const Enumerator *  enumerator )
                            :
                               m_name(name),
                               m_width(width),
                               m_start(start),
                               m_end(end),
                               m_type(RegisterFieldType_Enum),
                               m_enumerator(enumerator)
                            {
                            }

                        public:
                            inline Field( const std::string &   name,
                                          width_t               width,
                                          range_t               start,
                                          range_t               end,
                                          bool                  is_signed)
                            :
                               m_name(name),
                               m_width(width),
                               m_start(start),
                               m_end(end),
                               m_type(is_signed ? RegisterFieldType_Signed : RegisterFieldType_UnSigned),
                               m_enumerator(nullptr)
                            {
                            }
                    };

                public:
                    typedef std::vector<const Field*>       fields_collection_t;
                    typedef std::set<std::string>           dumps_collection_t;


                private:
                    const std::string               m_name;
                    const address_t                 m_address;
                    const register_access_flags_t   m_access;
                    const register_unit_flags_t     m_units;
                    const register_secure_t         m_secure;

                private:
                    fields_collection_t             m_fields;
                    dumps_collection_t              m_dumps;

                public:
                    inline const std::string &          name() const    { return m_name;    }
                    inline address_t                    address() const { return m_address; }
                    inline register_access_flags_t      access() const  { return m_access;  }
                    inline register_unit_flags_t        units() const   { return m_units;   }
                    inline register_secure_t            secure() const  { return m_secure;  }

                public:
                    inline const dumps_collection_t &   dumps() const   { return m_dumps;   }
                    inline const fields_collection_t &  fields() const  { return m_fields;  }

                private:
                    Register() :
                        m_address(0),
                        m_access(RegisterAccessFlags_None),
                        m_units(RegisterUnitFlags_None),
                        m_secure(RegisterSecureType_None)
                    {}

                public:
                    Register(const std::string &        name,
                             address_t                  address,
                             register_access_flags_t    access,
                             register_unit_flags_t      units,
                             register_secure_t          secure,
                             dumps_collection_t &       dumps)
                    :
                        m_name(name),
                        m_address(address),
                        m_access(access),
                        m_units(units),
                        m_secure(secure),
                        m_dumps(dumps)
                    {}

                public:
                    ~Register() {
                        release_container_data(m_fields);
                    }

                public:
                    inline const Field * add(const Field * field)
                    {
                        m_fields.push_back(field);
                        return field;
                    }
            };

        public:
            class Variant
            {
                private:
                    const Register &                m_reg;
                    const register_unit_flags_t     m_unit;

                public:
                    inline const Register &         reg() const      { return m_reg;     }
                    inline register_unit_flags_t    unit() const     { return m_unit;    }

                public:
                    Variant(const Register & reg, register_unit_flags_t unit)
                        : m_reg(reg), m_unit(unit)
                    {
                        ;
                    }

                    Variant(const Variant & variant)
                        : m_reg(variant.m_reg), m_unit(variant.m_unit)
                    {
                        ;
                    }
            };

        public:
            typedef std::map<std::string, const Enumerator*>                enum_collection_t;
            typedef std::map<uint16_t, const Register*>                     register_index_t;

            typedef std::vector<Variant>                                    variant_collection_t;
            typedef std::map<std::string, variant_collection_t>             variant_index_t;

            typedef uint16_t                                                value_t;

        private:
            const std::string           m_filename;
            dataset_type_t              m_type;
            const std::string           m_name;
            const Version               m_version;
            const std::string           m_release;

        private:
            enum_collection_t           m_enums;
            register_index_t            m_registers_index;
            mutable variant_index_t     m_variant_index;

        public:
            inline const std::string &              filename() const        { return m_filename;        }
            inline dataset_type_t                   type() const            { return m_type;            }
            inline const std::string &              name() const            { return m_name;            }
            inline const Version &                  version() const         { return m_version;         }
            inline const std::string &              release() const         { return m_release;         }
            inline const enum_collection_t &        enums() const           { return m_enums;           }
            inline const register_index_t &         registers() const       { return m_registers_index; }

        private:
            DataSet() : m_type(DataSetType_None), m_version(0,0) {}

        public:
            DataSet(const std::string & filename, dataset_type_t type, const std::string & name, const Version & version, const std::string & release)
                : m_filename(filename), m_type(type), m_name(name), m_version(version), m_release(release)
            {
            }

        public:
            ~DataSet() {
                release_container_data(m_enums);
                release_container_data(m_registers_index);
                m_variant_index.clear();
            }

        public:
            const Register * add(const Register * reg);
            const Enumerator * add(const Enumerator * enumerator);

        public:
            const variant_collection_t & variants(const std::list<std::string> & dumps) const;

    };
}
